from django.apps import apps as django_apps
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules


class Helper:
    def __init__(
        self,
        subject_identifier=None,
        subject_consent_model_cls=None,
        onschedule_model_name=None,
    ):
        self.subject_identifier = subject_identifier
        self.subject_consent_model_cls = subject_consent_model_cls or django_apps.get_model(
            "edc_visit_tracking.subjectconsent"
        )
        self.onschedule_model_name = (
            onschedule_model_name or "edc_visit_tracking.onscheduleone"
        )

    def consent_and_put_on_schedule(
        self,
        subject_identifier=None,
    ):
        subject_identifier = subject_identifier or self.subject_identifier
        subject_consent = self.subject_consent_model_cls.objects.create(
            subject_identifier=subject_identifier, consent_datetime=get_utcnow()
        )
        _, schedule = site_visit_schedules.get_by_onschedule_model(self.onschedule_model_name)
        schedule.put_on_schedule(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        return subject_consent



from ..exceptions import RelatedVisitFieldError
from .visit_model_mixin import VisitModelMixin


def get_related_visit_model_attr(model_cls) -> str:
    attrs = []
    for fld_cls in model_cls._meta.get_fields():
        if fld_cls.related_model is not None and issubclass(
            fld_cls.related_model, (VisitModelMixin,)
        ):
            attrs.append(fld_cls.name)
            break
    if len(attrs) > 1:
        raise RelatedVisitFieldError(
            f"More than one field is related to a visit model. Got {attrs}."
        )
    elif len(attrs) == 0:
        raise RelatedVisitFieldError(
            f"{model_cls} has no related visit model. "
            f"Expected the related visit model to be an instance "
            "of `VisitModelMixin`."
        )
    return attrs[0]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from ...exceptions import RelatedVisitFieldError
from ..utils import get_related_visit_model_attr
from ..visit_model_mixin import VisitModelMixin

if TYPE_CHECKING:
    from django.db.models import OneToOneField


class VisitMethodsModelMixin(models.Model):
    """A model mixin for CRFs and Requisitions to add methods to
    access the related visit model and its attributes.

    Used by VisitTrackingCrfModelMixin.
    """

    def __str__(self) -> str:
        return str(self.related_visit)

    def natural_key(self) -> tuple:
        return tuple(getattr(self, self.related_visit_model_attr()).natural_key())

    natural_key.dependencies = [settings.SUBJECT_VISIT_MODEL]

    @classmethod
    def related_visit_model_attr(cls) -> str:
        """Returns the field name for the visit model foreign key."""
        return get_related_visit_model_attr(cls)

    @classmethod
    def related_visit_field_cls(cls) -> OneToOneField | None:
        """Returns the field class of the visit foreign key attribute."""
        related_visit_field_cls = None
        for fld_cls in cls._meta.get_fields():
            if fld_cls.name == cls.related_visit_model_attr():
                related_visit_field_cls = fld_cls
                break
        if not related_visit_field_cls:
            raise RelatedVisitFieldError(f"Related visit field class not found. See {cls}.")
        return related_visit_field_cls

    @classmethod
    def related_visit_model_cls(cls) -> VisitModelMixin:
        """Returns the model class of the visit foreign key attribute."""
        related_model = None
        for fld_cls in cls._meta.get_fields():
            if fld_cls.name == cls.related_visit_model_attr():
                related_model = fld_cls.related_model
                break
        if not related_model:
            raise RelatedVisitFieldError(f"Related visit field class not found. See {cls}.")
        return related_model

    @classmethod
    def related_visit_model(cls) -> str:
        """Returns the name of the visit foreign key model in label_lower format"""
        return cls.related_visit_model_cls()._meta.label_lower

    @property
    def visit_code(self) -> str:
        return self.related_visit.visit_code

    @property
    def visit_code_sequence(self) -> int:
        return self.related_visit.visit_code_sequence

    @property
    def related_visit(self) -> VisitModelMixin:
        """Returns the model instance of the visit foreign key
        attribute.

        Note: doing this will cuase a RelatedObjectDoesNotExist exception:
            return getattr(self, self.related_visit_model_attr())
        RelatedObjectDoesNotExist cannot be imported since it is created
        at runtime.
        """
        related_visit = None
        for fld_cls in self._meta.get_fields():
            try:
                related_model = fld_cls.related_model
            except AttributeError:
                pass
            else:
                if related_model is not None and issubclass(related_model, (VisitModelMixin,)):
                    try:
                        related_visit = getattr(self, fld_cls.name)
                    except ObjectDoesNotExist:
                        pass
        return related_visit

    class Meta:
        abstract = True



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms

if TYPE_CHECKING:
    from edc_crf.crf_form_validator import CrfFormValidator
    from edc_crf.form_validator_mixins import CrfFormValidatorMixin
    from edc_crf.modelform_mixins import InlineCrfModelFormMixin

    from ..model_mixins import VisitModelMixin
    from ..modelform_mixins import VisitTrackingCrfModelFormMixin


def get_related_visit(
    modelform: (
        VisitTrackingCrfModelFormMixin
        | InlineCrfModelFormMixin
        | CrfFormValidator
        | CrfFormValidatorMixin
    ),
    related_visit_model_attr: str = None,
) -> VisitModelMixin | None:
    """Returns the related visit mode instance or None.

    Tries instance and cleaned data.
    """
    if related_visit_model_attr not in modelform.cleaned_data:
        related_visit = modelform.instance.related_visit
        if not related_visit:
            raise forms.ValidationError(f"Field `{related_visit_model_attr}` is required (2).")
    else:
        related_visit = modelform.cleaned_data.get(related_visit_model_attr)
    return related_visit
